
import javax.swing.*;
import java.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.event.*;
import java.awt.image.*;
import java.io.*;
import com.sun.image.codec.jpeg.*;

import javax.vecmath.*;


class Sphere {
    Sphere m_parent;
    double m_x, m_y, m_z;
    double m_r;
    double m_targetr; // last radius
    ArrayList<Sphere> m_children;
    int m_depth; // 0 for soma
    int m_cell;

    Vector3d m_force = new Vector3d();
    
    static int SECTION_SIZE = 13; // num spheres on a section
    static int MAX_DEPTH = SECTION_SIZE * 5; 
    // or min feature size???
    static int SPLIT_ORDER = 2;


    Sphere(double x, double y, double z, double r, Sphere parent) {
	m_x = x;
	m_y = y;
	m_z = z;
	m_r = r;
	m_parent = parent;	
	m_depth = 0;

	if (m_parent != null) { 
	    m_depth = m_parent.m_depth +1;
	}
	m_targetr = m_r;
    }


    void setGlia() {
	m_parent = this;
    }

    // glia are single spheres...
    boolean isGlia() {
	return m_parent==this;
    }

    void setForce(Vector3d f) {
	m_force = f;
    }

    Vector3d getForce() { return m_force; }

    void applyForce(double frac) {	
	m_force.scale(frac);
	double len = m_force.length();	
	if (len > 0.001) {
	    if (len > 1) {
		m_force.normalize();
	    }
	    m_x += m_force.x;
	    m_y += m_force.y;
	    m_z += m_force.z; 
	} 
    }

    void setCell(int cell) {
	m_cell = cell;
    }

    int getCell() { return m_cell; }

    void setTargetRadius(double r) {
	m_targetr = r;
    }
    
    int numChildren() {
	if (m_children==null) {
	    return 0;
	}
	return m_children.size();
    }
    
    void add(Sphere child) {
	if (m_children == null) {
	    m_children = new ArrayList<Sphere>();
	    m_children.add(child);	    
	}
    }

    boolean firstChild() {
	return (m_parent != null && m_parent.m_parent==null);
    }

    boolean onlyChild() {       
	return m_parent==null || m_parent.numChildren() == 1;
    }

    double targetRadius() {
	return m_targetr;
    }
    
    void addChildren() {
	m_children = new ArrayList<Sphere>();
	// get unit vector from parent.
	Vector3d dir = new Vector3d(1,0,0);
	
	if (m_parent != null) {
	    dir.set((m_x-m_parent.m_x) * (.5 + Math.random()) ,
		    (m_y-m_parent.m_y) * (.5 + Math.random()),
		    (m_z-m_parent.m_z) * (.5 + Math.random()));	   
	    dir.normalize();
	}

	int num = 1;

	int ssize = SECTION_SIZE;
	int sorder = SPLIT_ORDER;
	if (m_depth > 20) { 
	    ssize *= 3;
	    sorder += 1;
	}

	if (m_depth % ssize == 0) {
	    num = sorder;
	}

	// angle of children???
	double twist = 2*Math.PI / num;

	double childr = m_targetr;
	double initr = m_targetr;

	if (num > 1) {
	    childr = m_targetr / 1.5;
	}

	initr = childr * 0.25;


	AxisAngle4d twister = new AxisAngle4d(dir, twist);
	Matrix3d twistm = new Matrix3d();
	twistm.set(twister);
	double start = 0; // or existing twist.

	double theta = 0; // offset

	if (num > 1) {
	    theta = Math.PI / (6 + Math.random());
	}

	// get a root normal to dir
	Vector3d perp1 = new Vector3d(0,1,0);
	Vector3d perp2 = new Vector3d(0,0,1);
	Vector3d norm = new Vector3d();
	if (true || dir.dot(perp1) < dir.dot(perp2)) {
	    norm.cross(dir, perp1);
	    norm.normalize();
	}
	else {
	    norm.cross(dir, perp2);
	    norm.normalize();
	}
	
	Vector3d pos1 = new Vector3d(m_x, m_y, m_z);
	// go towards edge of sphere.
	dir.scale(m_r * Math.cos(theta));
	pos1.add(dir);
	norm.scale(Math.sin(theta) * m_r);
	

	for (int c=0; c<num; c++) {



	    Vector3d cpos = new Vector3d(pos1);
	    cpos.add(norm);

	    // start at half target...
	    Sphere child = new Sphere( cpos.x, cpos.y, cpos.z, initr, this );
	    child.setTargetRadius( childr );
	    m_children.add(child);

	    twistm.transform(norm);	    
	}	
    }
    
    // move all children by dir.
    void move(Vector3d dir) {
	m_x += dir.x;
	m_y += dir.y;
	m_z += dir.z;	
	for (int i=0; i<numChildren(); i++) {
	    m_children.get(i).move(dir);
	}
    }

    // recusively moves children after expanding by dist.
    void moveChildren(double dist) {
	for (int i=0; i<numChildren(); i++) {
	    Sphere child = m_children.get(i);
	    Vector3d childdir = new Vector3d( child.m_x - m_x, child.m_y - m_y, child.m_z - m_z);
	    childdir.normalize();
	    childdir.scale(dist);
	    child.move(childdir);
	}	
    }


    boolean update() {	
	boolean done = false;
	if (m_parent != null) {
	    if (m_r < m_targetr) {
		//	System.out.println("moving a sphere from "+m_r+" to "+m_targetr);
		double newr = m_r * 1.05;
		if (newr > m_targetr) newr = m_targetr;
		double dist = newr - m_r;
		m_r = newr;
		moveChildren(dist);
	    }
	}
	for (int i=0; i<numChildren(); i++) {
	    done |= m_children.get(i).update();
	}
	if (m_depth < MAX_DEPTH && m_children==null) {
	    addChildren();
	}
	if (m_depth == MAX_DEPTH) {
	    System.out.println("update returning true");
	    return true;
	}
	return done;
    }

    static Vector3d m_cen = new Vector3d();
    static Ellipse2D.Double m_shape = new Ellipse2D.Double();
    void plot(Graphics2D g, Matrix3d rot) {
	// System.out.println("plotting a sphere at "+m_x+" "+m_y+" " +m_r+ "dep "+m_depth);
	m_cen.set( m_x, m_y, m_z);
	rot.transform(m_cen);
	m_shape.x = 200 + m_cen.x-m_r;
	m_shape.y = 200 + m_cen.y-m_r;
	m_shape.width = m_r*2;
	m_shape.height = m_r*2;

	//g.draw(m_shape);
	g.fill(m_shape);

	for (int i=0; i<numChildren(); i++) {
	    m_children.get(i).plot(g, rot);
	}
    }
    
    void addToList(LinkedList list, int cell) {
	m_cell = cell;
	list.add(this);
	for (int i=0; i<numChildren(); i++) {
	    m_children.get(i).addToList(list, cell);
	}
    }


    void dump(FileWriter fw) throws IOException {
	fw.write("sphere "+m_x+" "+m_y+" "+m_z+" "+m_r+"\n");
	for (int i=0; i<numChildren(); i++) {
	    m_children.get(i).dump(fw);
	}	
    }

}

class Spheres {

    ArrayList<Sphere> m_somas;
    ArrayList<Sphere> m_glia;

    Spheres() {
	m_somas = new ArrayList<Sphere>();
	m_glia = new ArrayList<Sphere>();

	int num = 3;
	int start = -num/2;

	for (int x=start; x<start+num; x++) {
	    for (int y=start; y<start+num ; y++) { // ***
		Sphere root = new Sphere( 0,y*60,x*60,30, null);
		Sphere dend = new Sphere( 30,root.m_y,x*60,6, root);
		root.add(dend);
		m_somas.add(root);

		root = new Sphere( 400,y*60,x*60,30, null);
	        dend = new Sphere( 370,y*60,x*60,6, root);
		root.add(dend);
		m_somas.add(root);
		
	    }
	}
	
	if (false) {

	for (int x = 0; x < 400; x += 50) {
	    for (int y = -150; y < 150; y += 50) {
		for (int z = -150; z < 150; z += 50) {
		    Sphere g = new Sphere(x,y,z,10,null);
		    g.setGlia();
		    m_glia.add(g);		    
		}
	    }	    
	}
	}

    }	


    // dump out to "spheres.txt" file.
    void dump() {
	
	try {
	FileWriter fw = new FileWriter("spheres.txt");
       	for (int i=0; i<m_somas.size(); i++) {	  
	    int ci = (i % 7) + 1;
	    int r = (ci & 1) ==0 ? 200 : 255;
	    int g = (ci & 2) ==0 ? 200 : 255;
	    int b = (ci & 4) ==0 ? 200 : 255;
	    
	    fw.write("surface "+((double)r/255.0)+" "+((double)g/255.0)+" "+((double)b/255.0)+" 0.3 0.6 0.4 40.0 0.2 0.0 0\n");
	    m_somas.get(i).dump(fw);
	}
	fw.close();

	} catch (IOException ie) {
	    System.out.println("prob "+ie);
	}
	

	
    }

    LinkedList m_boxes[][][];
    int m_nb;
    double m_bsize = 20;
    double m_bmin = -400;
    double m_bmax = 400;
    LinkedList<Sphere> m_all;

    // apply gravity to system
    void clumpInit() {

	System.out.println("Clumping...");

	// get spheres
	m_all = new LinkedList<Sphere>();
	
       	for (int i=0; i<m_somas.size(); i++) {	    
	    m_somas.get(i).addToList(m_all, i);
	}

       	for (int i=0; i<m_glia.size(); i++) {	    
	    m_glia.get(i).addToList(m_all, -1-i);
	}

	
	m_nb = (int) ((m_bmax - m_bmin) / m_bsize);

	// make boxes
	m_boxes = new LinkedList[m_nb][m_nb][m_nb];
	

	// put in boxes
	Iterator<Sphere> iter = m_all.iterator();
	int maxInBox = 0;
	while (iter.hasNext()) {
	    Sphere s = iter.next();

	    int bx = (int)((s.m_x - m_bmin) / m_bsize);
	    int by = (int)((s.m_y - m_bmin) / m_bsize);
	    int bz = (int)((s.m_z - m_bmin) / m_bsize);

	    if (bx < 0) bx = 0;
	    if (by < 0) by = 0;
	    if (bz < 0) bz = 0;
	    if (bx >= m_nb) bx = m_nb-1;
	    if (by >= m_nb) by = m_nb-1;
	    if (bz >= m_nb) bz = m_nb-1;

	    // add to box...
	    LinkedList<Sphere> box = m_boxes[bx][by][bz];
	    if (box == null) {
		box = new LinkedList<Sphere>();
		m_boxes[bx][by][bz] = box;
	    }
	    box.add(s);	 
	    int n = box.size();
	    if (n > maxInBox) {
		maxInBox = n;
	    }
	}	
	System.out.println("Put into boxes, max num is "+maxInBox+" total is "+m_all.size());	
    }

    void clumpStep() {
	double maxForce = 0.00001;

	// compare each sphere with its neighbours.
	for (int bx = 0; bx < m_nb; bx ++) {
	    for (int by = 0; by < m_nb; by ++) {
		for (int bz = 0; bz < m_nb; bz ++) {
		    LinkedList<Sphere> box = m_boxes[bx][by][bz];
		    if (box != null) {
			Iterator<Sphere> iter = box.iterator();
			while (iter.hasNext()) {
			    Sphere s = iter.next();
			    Vector3d f = s.getForce();
			    f.set(0,0,0);
			    calcForce(s, f, bx, by, bz);
			    
			    double mag = f.length();
			    if (mag > maxForce) maxForce = mag;
			    
			    // s.apply(force);  // not yet...
			}
		    }
		}
	    }
	}	
	
	if (maxForce > 1000) { maxForce = 1000; }

	double scale = 1 / maxForce;

	System.out.println("scale is "+scale+" maxForce is "+maxForce);
	if (maxForce > 20) {
	    // apply scaled forces...
	    Iterator<Sphere> iter = m_all.iterator();
	    while (iter.hasNext()) {
		Sphere s = iter.next();
		s.applyForce(scale);
	    }
	}
	else {
	    System.out.println("Reached stable state, maxforce is "+maxForce);
	}
    }
    
    boolean checkbounds(int x) {
	return x>=0 && x < m_nb;
    }

    boolean checkbounds(int x, int y, int z) {
	return checkbounds(x) && checkbounds(y) && checkbounds(z);
    }

    void calcForce(Sphere s, Vector3d force, int bx, int by, int bz) {
	for (int dx=-1; dx<=1; dx ++) {
	    for (int dy=-1; dy<=1; dy ++) {
		for (int dz=-1; dz<=1; dz ++) {
		    int x = bx + dx;
		    int y = by + dy;
		    int z = bz + dz;
		    if (checkbounds(x,y,z)) {
			calcBoxForce(s, force, x,y,z);
		    }		    
		}
	    }
	}	
    }

    void calcBoxForce(Sphere s, Vector3d force, int bx, int by, int bz) {
	LinkedList<Sphere> list = (LinkedList<Sphere>)m_boxes[bx][by][bz];
	if (list==null) return;
	Iterator<Sphere> iter = list.iterator();
	while (iter.hasNext()) {
	    Sphere o = iter.next();
	    updateForce(s, o, force);
	}
    }

    // calc force between two spheres in dir (o - s) + attractive.
    Vector3d m_scratch = new Vector3d();
    double m_G = 10; // if dist of 10, force of 1, 
    double m_D = 0.001; // density

    void updateForce(Sphere s, Sphere o, Vector3d f) {
	if (s == o) return; 
	if (s.m_parent == null) return; // no force on soma.
	boolean self = (o.m_cell == s.m_cell);

	m_scratch.set(o.m_x - s.m_x, o.m_y - s.m_y, o.m_z-s.m_z);

	double dist = m_scratch.length();
	double tgtdist = o.m_r + s.m_r;

	boolean attached = false;
	
	if (self) {
	    if ((o == s.m_parent) || (s.m_parent == o)) {
		// ok, attached - aim for larger of two radii
		tgtdist = o.m_r;
		if (s.m_r > tgtdist) {
		    tgtdist = s.m_r;
		}		
	    }
	    else {
		if (dist <= tgtdist-1) {
		    m_scratch.normalize();

		    // System.out.println("repelling "+(dist-tgtdist));

		    m_scratch.scale((dist-tgtdist) * 1); 
		    f.add(m_scratch);
		}		
		return; // no forces to self.... apart from overlap...
	    }
	}
	
	
	if (dist <= tgtdist) {
	    // overlapping !!!  magnitude = -10;
	    m_scratch.normalize();
	    m_scratch.scale((dist-tgtdist )* 100); 
	    f.add(m_scratch);
	}
	else {
	    double excess = dist - tgtdist;
	    
	    if (excess > 2 && excess < m_bsize) {				
		double mag;

		if (self) {
		    mag = excess * 100; 
		} // forces bigger for self.
		else {
		    double m1 = m_D * s.m_r * s.m_r * s.m_r;
		    double m2 = m_D * o.m_r * o.m_r * o.m_r; // e.g. 1000 for 10, 30 for 3, 125000 for 50		
		    // mag = excess;
		    m1 = 1;
		    m2 = 1; // ignore mass
		    // mag = m_bsize - excess;
		    mag = m_G * m1 * m2 / excess; // or inv square law...
		}
		m_scratch.normalize();
		m_scratch.scale(mag);
		f.add(m_scratch);		    
	    }	    
	}	
    }


    boolean update() {
	boolean done = false;
	for (int i=0; i<m_somas.size(); i++) {
	    done |= m_somas.get(i).update();
	}
	return done;
    }

    void plot(Graphics2D g, int w, int h, Matrix3d rot) {
	System.out.println("plotting spheres");
	g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
	for (int i=0; i<m_somas.size(); i++) {
	    int ci = (i % 7) + 1;
	    
	    g.setColor( new Color( (ci & 1) ==0 ? 200 : 255,  (ci & 2) ==0 ? 200 : 255, (ci & 4) ==0 ? 200: 255, 120)); 


	    //	    g.setColor( new Color( 200,250,250,170));
	    m_somas.get(i).plot(g, rot);
	}		

	for (int i=0; i<m_glia.size(); i++) {
	    g.setColor( new Color( 50,250,50,20));
	    m_glia.get(i).plot(g, rot);
	}

    }
}


class GrowthPanel extends JPanel {

    Spheres m_spheres;

    double m_xrot, m_yrot, m_zrot;
    int m_but;

    public GrowthPanel() {

	m_spheres = new Spheres();

	m_xrot = 0.2;
	m_yrot = 0.2;
	m_zrot = 0.2;

	setPreferredSize(new Dimension(600,400));
	
	addMouseListener(new MouseInputAdapter() {
		int x1,y1,x2,y2;
		boolean m_zooming = false;
		public void mousePressed(MouseEvent me) {
		    m_but = me.getButton();
		}
		public void mouseReleased(MouseEvent me) {
		    
		    repaint();
		}
	    });


	addMouseMotionListener(new MouseMotionAdapter() {
		int x1,y1,x2,y2;
		boolean m_zooming = false;
		public void mouseDragged(MouseEvent me) {
		    x1=me.getX();
		    y1=me.getY();
		    
		    if (m_but == MouseEvent.BUTTON1) {
			// could change the x / y / z rotation.
			m_xrot = (double) x1 / 150.0;
			m_yrot = (double) y1 / 150.0;
			// step();
			repaint();
			
		    }
		    else if (m_but == MouseEvent.BUTTON3) {
			m_xrot = (double) x1 / 150.0;
			m_zrot = (double) y1 / 150.0;
			// step();
			repaint();
		    }
		}
	    });
	Anim anim = new Anim(this);
	anim.start();
    }
	
    public void reset() {
	repaint();
    }

    public void top() {
	m_xrot = 0.0;
	m_yrot = 0.0;
	m_zrot = 0.0;
	repaint();
    }

    public void left() {
	m_yrot += Math.PI / 36;
	repaint();
    }

    public void front() {
	m_xrot += Math.PI / 36;
	repaint();
    }

    static Color m_invis = new Color(0,0,0,0);

    boolean step() {
	boolean done = m_spheres.update();
	repaint();
	return done;
    }

    void clumpStep() {
	m_spheres.clumpStep();
    }

    void dump() {
	m_spheres.dump();
    }

    void clumpInit() {
	m_spheres.clumpInit();
    }

    Matrix3d getRotMatrix( double xrot, double yrot, double zrot ) {
	Matrix3d rx = new Matrix3d(); rx.rotX(xrot);
	Matrix3d ry = new Matrix3d(); ry.rotY(yrot);
	Matrix3d rz = new Matrix3d(); rz.rotZ(zrot);	
	rx.mul(ry);
	rx.mul(rz);
	return rx;
    }
    
    public void paint(Graphics gg) {
	
	//	System.out.println("repainting "+m_xrot+" "+m_yrot+" "+m_zrot);
	int w = getWidth();
	int h = getHeight();

	gg.setColor(Color.gray);
	gg.fillRect(0,0,w,h);

	Graphics2D g2d = (Graphics2D)gg;

	Matrix3d rot = getRotMatrix(m_xrot, m_yrot, m_zrot);

	m_spheres.plot(g2d, w, h, rot);
	// plot the stuff...

    }    
}

class Anim extends Thread {
    GrowthPanel m_panel;
    Anim(GrowthPanel panel) {
	m_panel = panel;
    }
    public void run() {
	try {
	    boolean done = false;
	    for (int i=0; !done && i<100000; i++) {
		done = m_panel.step();		
		
		Thread.sleep(133);
	    }	    
	} catch (Exception ex) {

	}

	m_panel.clumpInit();
	System.out.println("finished animating");
    }
}



public class Growth {
    GrowthPanel m_panel;

    public Growth() {
    }

    void doit() {
	JFrame frame = new JFrame("Growth test");

	frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);


	m_panel = new GrowthPanel();

	JPanel main = new JPanel(new BorderLayout());
	JPanel buts = new JPanel();
	JButton reset = new JButton("Reset");
	JButton top = new JButton("Top");
	JButton left = new JButton("Left");
	JButton front = new JButton("Front");
	JButton step = new JButton("Step");
	JButton clump = new JButton("Clump");
	JButton dump = new JButton("Dump to file");


	buts.add(reset);
	reset.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.reset();
		}
	    });

	buts.add(top);
	top.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.top();
		}
	    });

	buts.add(left);
	left.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.left();
		}
	    });

	buts.add(front);
	front.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.front();
		}
	    });

	buts.add(step);
	step.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    for (int i=0; i<1; i++)
			m_panel.step();
		    m_panel.repaint();
		}
	    });

	buts.add(clump);
	clump.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    for (int i=0; i<5; i++) {
			m_panel.clumpStep();
		    }
		    m_panel.repaint();
		}
	    });

	buts.add(dump);
	dump.addActionListener( new AbstractAction() {
		public void actionPerformed(ActionEvent ae) {
		    m_panel.dump();
		}
	    });


	frame.getContentPane().add(main);
	main.add(m_panel);
	main.add(buts, BorderLayout.SOUTH);
	frame.pack();
	frame.show();
    }
    
    
    public static void main(String[] args) {
	Growth m = new Growth();
	m.doit();
    }
    
}

